package gecco2015.utils;

import jmetal.core.Solution;
import jmetal.core.Variable;
import jmetal.util.JMException;
import weka.classifiers.Classifier;
import weka.core.Attribute;
import weka.core.FastVector;
import weka.core.Instance;
import weka.core.Instances;

import java.util.ArrayList;

public class SurrogateModel {

    Classifier model;
    Instances trSet;
    String modelOptions;
    String modelClass;
    double[] uB;
    double[] lB;

    public SurrogateModel(String modelClass, String modelParameters) {
        this.modelOptions = modelParameters;
        this.modelClass = modelClass;
        try {
            model = (Classifier)Class.forName(modelClass).newInstance();
            model.setOptions(modelOptions.split(" "));
        } catch (InstantiationException | IllegalAccessException | ClassNotFoundException e) {
            e.printStackTrace();
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    public double value(double[] point) {
        //for (int i = 0; i < point.length; i++) {
        //    point[i] = (point[i] - lB[i])/(uB[i] - lB[i]);
        //}

        Instance i1 = new Instance(1, point);
        i1.setDataset(trSet);
        try {
            return model.classifyInstance(i1);
        } catch (Exception e) {
            //e.printStackTrace();
        }

        return 0.0;
    }

    public double value(Solution sol) {
        return value(solutionToDoubleArray(sol));
    }

    public double[] numericalGradient(double[] point) {
        double value = value(point);

        double[] grad = new double[point.length];

        for (int i = 0; i < grad.length; i++) {
            double h = 0.0000001 * point[i];
            double[] xph = point.clone();
            xph[i] += h;
            double dx = xph[i] - point[i];
            grad[i] = (value(xph) - value) / dx;
            if (Double.isNaN(grad[i])) {
                grad[i] = 0;
            }
        }

        return grad;
    }

    public void train(ArrayList<Solution> trainingSet, ArrayList<Double> targetValues) {

        int vars = trainingSet.get(0).getDecisionVariables().length;

        FastVector attributes = new FastVector(vars + 1);
        lB = new double[vars];
        uB = new double[vars];


        /*try {
            for (int i = 0; i < vars; i++) {
                lB[i] = trainingSet.get(0).getDecisionVariables()[i].getLowerBound();
                uB[i] = trainingSet.get(0).getDecisionVariables()[i].getUpperBound();
            }
        } catch (JMException e) {
            e.printStackTrace();
        }*/

        for (int i = 0; i < vars + 1; i++) {
            attributes.addElement(new Attribute("a" + i));
        }

        trSet = new Instances("train", attributes, 0);
        trSet.setClassIndex(vars);

        for (int j = 0; j < trainingSet.size(); j++) {
            Solution s = trainingSet.get(j);
            double[] x = new double[vars + 1];
            for (int i = 0; i < vars; i++) {
                try {
                    x[i] = s.getDecisionVariables()[i].getValue();
                    //x[i] = (x[i] - lB[i])/(uB[i] - lB[i]);
                } catch (JMException e) {
                    e.printStackTrace();
                }
            }
            Instance i1 = new Instance(1, x);
            i1.setDataset(trSet);
            i1.setClassValue(targetValues.get(j));
            trSet.add(i1);
        }

        try {
            model.buildClassifier(trSet);
        } catch (Exception e) {
            e.printStackTrace();
        }
    }

    private double[] solutionToDoubleArray(Solution sol) {
        Variable[] vars = sol.getDecisionVariables();
        double[] dbls = new double[vars.length];
        try {
            for (int i = 0; i < vars.length; i++) {
                dbls[i] = vars[i].getValue();
            }
        } catch (JMException e) {
            e.printStackTrace();
        }
        return dbls;
    }
}
